import 'dart:developer';
import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:nui_core/code/nui_code_util.dart';
import 'package:nui_core/data/nui_data_util.dart';
import 'package:nui_core/datetime/nui_dt_util.dart';
import 'package:nui_core/string/nui_string_util.dart';
import 'package:nui_form/nuif_form.dart';
import 'package:nui_form/nuif_form_component.dart';

typedef OnTextEditingComplete = void Function();
typedef OnTextSubmitted = void Function(String value); //Clicked on next/done
typedef TextFormatter = String Function(String value, NUIFTextInputType inputType, NUIFTextFieldHandler handler);
typedef ContentCheck = bool Function(String value);

enum NUIFTextDesign{
  NORMAL,
  BORDER,
  NONE
}

enum NUIFTextInputType{
  PASSWORD_NEW,
  PASSWORD,
  EMAIL,
  NAME,
  NUMBER,
  PHONE,
  DECIMAL,
  SENTENCE,
  DATE,
  DOB,
  TIME,
  DATETIME,
  MULTILINE,
  MALAYSIAN_IC,
  MALAYSIAN_PHONE
}

extension NUIFTextInputExt on NUIFTextInputType{
  static const String passwordNew = "passwordNew";
  static const String password = "password";
  static const String email = "email";
  static const String date = "date";
  static const String dob = "dob";
  static const String dateTime = "dateTime";
  static const String decimal = "decimal";
  static const String number = "number";
  static const String phone = "phone";
  static const String sentence = "sentence";
  static const String time = "time";
  static const String multiline = "multiline";
  static const String malaysianPhone = "phone_MY";
  static const String malaysianIC = "ic_MY";
  static const String name = "name";

  String value(){
    switch(this){
      case NUIFTextInputType.PASSWORD_NEW : return passwordNew;
      case NUIFTextInputType.PASSWORD : return password;
      case NUIFTextInputType.EMAIL : return email;
      case NUIFTextInputType.DECIMAL : return decimal;
      case NUIFTextInputType.DOB : return dob;
      case NUIFTextInputType.NUMBER : return number;
      case NUIFTextInputType.PHONE : return phone;
      case NUIFTextInputType.SENTENCE : return sentence;
      case NUIFTextInputType.DATE : return date;
      case NUIFTextInputType.TIME : return time;
      case NUIFTextInputType.DATETIME : return dateTime;
      case NUIFTextInputType.MULTILINE : return multiline;
      case NUIFTextInputType.MALAYSIAN_PHONE : return malaysianPhone;
      case NUIFTextInputType.MALAYSIAN_IC : return malaysianIC;
      case NUIFTextInputType.NAME : return name;
      default: return "unknown";
    }
  }

  bool isPasswordInput(){
    return this == NUIFTextInputType.PASSWORD || this == NUIFTextInputType.PASSWORD_NEW;
  }

  static NUIFTextInputType parse(String value){
    switch(value){
      case passwordNew: return NUIFTextInputType.PASSWORD_NEW;
      case password: return NUIFTextInputType.PASSWORD;
      case email: return NUIFTextInputType.EMAIL;
      case date: return NUIFTextInputType.DATE;
      case decimal: return NUIFTextInputType.DECIMAL;
      case number: return NUIFTextInputType.NUMBER;
      case dob: return NUIFTextInputType.DOB;
      case phone: return NUIFTextInputType.PHONE;
      case sentence: return NUIFTextInputType.SENTENCE;
      case date: return NUIFTextInputType.DATE;
      case time: return NUIFTextInputType.TIME;
      case dateTime: return NUIFTextInputType.DATETIME;
      case multiline: return NUIFTextInputType.MULTILINE;
      case malaysianPhone: return NUIFTextInputType.MALAYSIAN_PHONE;
      case malaysianIC: return NUIFTextInputType.MALAYSIAN_IC;
      case name: return NUIFTextInputType.NAME;
      default: return null;
    }
  }
}

class NUIFTextStyle{
  NUIFTextStyle({
    this.border,
    this.disabledBorder,
    this.enabledBorder,
    this.focusedBorder,
    this.errorBorder,
    this.errorFocusedBorder,
    this.contentPadding,
    this.helperStyle,
    this.design,
    this.textStyle,
    this.hintStyle,
    this.labelStyle,
    this.errorLabelStyle,
    this.errorHelperStyle,
    this.cursor,
    this.fillColor,
    this.dense: false
  });
  OutlineInputBorder border;
  OutlineInputBorder disabledBorder;
  OutlineInputBorder enabledBorder;
  OutlineInputBorder focusedBorder;
  OutlineInputBorder errorBorder;
  OutlineInputBorder errorFocusedBorder;
  Color fillColor;
  EdgeInsets contentPadding;
  TextStyle helperStyle;
  TextStyle errorHelperStyle;
  NUIFTextDesign design;
  TextStyle hintStyle;
  TextStyle labelStyle;
  TextStyle errorLabelStyle;
  TextStyle textStyle;
  NUIFTextCursor cursor;
  bool dense;
}

class NUIFTextItem extends NUIFormComponent with NUIFValidatable{
  NUIFTextItem({
    @required this.controller,
    @required String inpId,
    double width,
    double height,
    double weight,
    bool enabled: true,
    bool compulsory: true,
    bool show: true,
    this.text: "",
    this.autoFocus: false,
    this.focusNode,
    this.nextFocusNode,
    this.onChanged,
    this.onSubmitted,
    this.onEditingComplete,
    this.inputType:TextInputType.text,
    this.inputAction:TextInputAction.done,
    this.autoCorrect:false,
    this.textCapitalization:TextCapitalization.words,
    this.textAlign: TextAlign.start,
    this.style,
    this.maxLength,
    this.minLength,
    this.maxLine,
    this.minLine,
    this.obscure:false,
    this.hint,
    this.label,
    this.helperText,
    this.icon,
    this.prefix,
    this.prefixText,
    this.prefixIcon,
    this.suffix,
    this.suffixText,
    this.suffixIcon,
    this.regExp,
    this.maskedInputType: NUIFTextInputType.SENTENCE,
    this.contentCheck,
    this.validateOnChange: false,
    this.extraFormat,
    this.postTextFormatter,
    this.blacklisting,
    this.whitelisting,
    this.additionalFormat,
    this.firstDate,
    this.lastDate,
    this.suffixOnTap,
    this.iconPasswordHide,
    this.iconPasswordVisible,
    this.validateOnEditComplete: false,
    this.validateOnSubmit: false,
    this.inputFormatters,
    Map<TextValidation, String> errorMessages,
    Alignment align,
    EdgeInsets margin,
  }) : super(inpId: inpId, inpType:ComponentType.TEXT, width: width, height : height, weight: weight, enabled: enabled, margin: margin, align: align, compulsory: compulsory, show: show) {
    this.errorMessages = errorMessages ?? Map();
    final defaultMessage = NUIFValidatable.defaultErrorMessages();
    for(var key in defaultMessage.keys){
      final dftMessage = defaultMessage[key];
      this.errorMessages[key] = this.errorMessages[key] ?? dftMessage;
    }
  }

  TextEditingController controller;
  String text;
  bool autoFocus;
  FocusNode focusNode;
  FocusNode nextFocusNode;
  OnTextSubmitted onChanged;
  OnTextEditingComplete onEditingComplete;
  OnTextSubmitted onSubmitted;
  TextFormatter postTextFormatter;
  TextInputType inputType;
  Function suffixOnTap;
  TextInputAction inputAction;
  bool autoCorrect;
  TextCapitalization textCapitalization;
  TextAlign textAlign;
  int maxLength;
  int minLength;
  int maxLine;
  int minLine;
  bool obscure;
  String extraFormat;
  String hint;
  String label;
  String helperText;
  Widget icon;
  Widget iconPasswordVisible;
  Widget iconPasswordHide;
  Widget prefix;
  String prefixText;
  Widget prefixIcon;
  Widget suffix;
  String suffixText;
  Widget suffixIcon;
  String regExp;
  NUIFTextStyle style;
  ContentCheck contentCheck;
  NUIFTextInputType maskedInputType;
  String errorHelperText;
  bool validateOnChange;
  bool validateOnSubmit;
  bool validateOnEditComplete;
  String additionalFormat;
  DateTime firstDate;
  DateTime lastDate;
  RegExp blacklisting; //Disable certain chars from the keyboard
  RegExp whitelisting; //Enable only specified chars from the keyboard
  List<TextInputFormatter> inputFormatters;

  void addAllTextChangeListener(AllTextChangeListener listener, NUIFTextFieldHandler handler){
    if(controller != null){
      controller.addListener(() {
        if(onChanged != null){
          onChanged(controller.text);
        }
        if(postTextFormatter != null && maskedInputType != null){
          postTextFormatter(controller.text, maskedInputType, handler);
//          if(!match(newValue, controller.text)){
//            controller.text = newValue;
//          }
        }
        updateValue(controller.text);
        if(listener != null) {
          listener(inpId, controller.text);
        }
      });
    }
  }

  void updateValue(String text){
    NUIFormValue value = formValue ?? NUIFormValue(inpId: inpId);
    value.value = NUIFormSelection(value: text, code: text);
    this.formValue = value;
  }

  @override
  bool validate(bool quiet){
    final curValue = formValue?.value?.value;
    log("NUIFText -> Validating text of id : $inpId, current value : $curValue");

    bool validatedAll = false;
    if(compulsory == true && curValue.isNullOrEmpty()){
      failed = true;
      formValue?.validated = !failed;
      errorHelperText = errorMessages[TextValidation.NOT_FILLED];
      validatedAll = false;
    }
    else {
      if (compulsory && minLength != null && curValue.length < minLength) {
        failed = true;
        formValue?.validated = !failed;
        errorHelperText = errorMessages[TextValidation.LESS_THAN_MIN_LENGTH];
        validatedAll = false;
      }
      else if(compulsory && maxLength != null && curValue.length > maxLength) {
        failed = true;
        formValue?.validated = !failed;
        errorHelperText = errorMessages[TextValidation.EXCEED_MAX_LENGTH];
        validatedAll = false;
      }
      else {
        // bool regExpSuccess;
        // if(maskedInputType == NUIFTextInputType.EMAIL){
        //   logNUI("NUIFText", "Validating email regexp for value : $curValue");
        //   regExpSuccess = NUIStringUtil.validateEmailPattern(curValue);
        // }
        // else {
        final regExpSuccess = NUIStringUtil.validateStringPattern(curValue, regExp);
        // }

        var validateSuccess = regExpSuccess;
        if(contentCheck != null){
          //Customized checking
          validateSuccess = contentCheck(curValue) && validateSuccess;
        }
        if(!regExpSuccess){
          errorHelperText = errorMessages[TextValidation.REG_EXPRESSION_FAILED];
        }
        else {
          errorHelperText = errorMessages[TextValidation.CONTENT_CHECK_FAILED];
        }
        failed = !validateSuccess;
        validated = validateSuccess;

        validatedAll = validateSuccess;
      }
    }

    if(!quiet && stateGetter != null){
      stateGetter().update(this);
    }

    return validatedAll;
  }

  @override
  Widget toWidget({AllButtonClickListener allButtonClickListener, AllTextChangeListener allTextChangeListener, AllItemSelectListener allItemSelectListener, Key key}) {
    return NUIFTextField(item: this, allTextChangeListener: allTextChangeListener, key: key);
  }

  @override
  void extractValue(Map<String, NUIFormValue> map){
    map[inpId] = formValue;
  }

  OutlineInputBorder getBorder(){
    return style?.border ?? OutlineInputBorder();
  }
  OutlineInputBorder getDisabledBorder(){
    return style?.disabledBorder ?? style?.border ?? OutlineInputBorder();
  }
  OutlineInputBorder getEnabledBorder(){
    return style?.enabledBorder ?? style?.border ?? OutlineInputBorder();
  }
  OutlineInputBorder getFocusedBorder(){
    return style?.focusedBorder ?? style?.border ?? OutlineInputBorder();
  }
  OutlineInputBorder getErrorBorder(){
    return style?.errorBorder ?? style?.border ?? OutlineInputBorder();
  }
  OutlineInputBorder getErrorFocusedBorder(){
    return style?.errorFocusedBorder ?? style?.border ?? OutlineInputBorder();
  }

  @override
  void fillInValue(Map<String, NUIFormValue> values) {
    logNUI("NUIFText", "Calling fill in values: $values, item id : $inpId");

    final formValue = values[inpId];
    if(formValue == null) return;

    text = formValue.value?.value;
    updateValue(text);
    if(stateGetter != null){
      stateGetter().update(this);
    }
  }
}

class NUIFTextCursor{
  NUIFTextCursor({this.color, this.cursorRadius, this.cursorWidth});
  Color color;
  Radius cursorRadius;
  double cursorWidth;
}

class NUIFTextField extends StatefulWidget {
  NUIFTextField({@required this.item, this.allTextChangeListener, Key key}): super(key: key);
  final NUIFTextItem item;
  final AllTextChangeListener allTextChangeListener;
  @override
  _NUIFTextFieldState createState() => _NUIFTextFieldState();
}

typedef NUIFTextFieldUpdateText = void Function(String text);
class NUIFTextFieldHandler{
  NUIFTextFieldHandler._builder();
  Function _updateState;
  String _newText;
  NUIFTextFieldUpdateText _updateText;
  void updateState(){
    if(_updateState != null) _updateState();
  }
  void updateText(String text){
    if(_updateText != null) _updateText(text);
  }
  String getNewText(){
    return _newText ?? "";
  }
}

class _NUIFTextFieldState extends NUIFState<NUIFTextField, NUIFTextItem> {
  bool _passwordToggle = false;
  final focusNode = FocusNode();
  String _newText = "";
  NUIFTextFieldHandler _handler;

  @override
  void initState(){
    this.item = widget.item;
    super.initState();

    _handler = NUIFTextFieldHandler._builder();
    _handler._updateText = (text){
      if(item.controller != null && item.controller.text.length < _newText.length) { // handling backspace in keyboard
        _newText = item.controller.text;
      }
      else {
        if (!match(text, _newText)) {
          _newText = text;
          _updateTextSelection(_newText, item.controller);
        }
      }
      _handler._newText = _newText;
    };
    _handler._updateState = (){
      setState(() {

      });
    };

    item.addAllTextChangeListener(widget.allTextChangeListener, _handler);
    focusNode.addListener(() {
      if(item.validateOnChange) {
        item.validate(false);
      }
      else if(item.validateOnSubmit == true || item.validateOnEditComplete == true){
        logNUI("NUIFtext", "Validate on submit and edit complete, current text : ${item.controller.text}");
        if(!isNullOrEmpty(item.controller.text)){
          item.validate(false);
        }
      }
    });
  }

  void _updateTextSelection(String value, TextEditingController controller){
    if(controller == null || isNullOrEmpty(value)) return;
    controller.text = value;
    logNUI("NUIFText", "Setting selection with value : $value");
    if(value?.length != null && value.length < 0){
      logNUI("NUIFText", "Invalid text selection");
    }
    else {
      controller.selection = new TextSelection(
          baseOffset: value?.length ?? 0,
          extentOffset: value?.length ?? 0
      );
    }
  }

  Widget _bindWithTextItem(NUIFTextItem item){
    InputDecoration decoration;
    bool disableEdit = false;

    Function onWholeFieldTapAction;

    if(item.style?.design == NUIFTextDesign.NORMAL || item.style?.design == NUIFTextDesign.BORDER){
      Widget suffixIcon;
      if(item.maskedInputType?.isPasswordInput() == true){
        suffixIcon = IconButton(icon : (_passwordToggle != false ? item.iconPasswordVisible : item.iconPasswordHide) ?? item.suffixIcon ?? Icon(Icons.remove_red_eye, color: item.style.labelStyle.color), onPressed: (){
          setState(() {
            FocusScope.of(context).unfocus();
            _passwordToggle = !_passwordToggle;
          });
        });
      }
      else if(item.maskedInputType == NUIFTextInputType.DATE){
        disableEdit = true;
        onWholeFieldTapAction = (){
          final selectedDate = NUIDTUtil.stringToDate(item.text ?? item.formValue?.value?.value);
          NUIDTUtil.dateSelect(context, selectedDate: selectedDate, selectListener: (date){
            if(date != null){
              final dateValue = dateToString(date, format: item.additionalFormat ?? NUIDTUtil.FORMAT_yyyy_MM_dd);
              item.text = dateValue;
              item.updateValue(dateValue);
              update(item);
            }
          }, firstDate: item.firstDate, lastDate: item.lastDate);
        };
        suffixIcon = IconButton(icon : item.suffixIcon ?? Icon(Icons.date_range, color: item.style.labelStyle.color), onPressed: (){
          final selectedDate = NUIDTUtil.stringToDate(item.text ?? item.formValue?.value?.value);
          NUIDTUtil.dateSelect(context, selectedDate: selectedDate, selectListener: (date){
            if(date != null){
              final dateValue = dateToString(date, format: item.additionalFormat ?? NUIDTUtil.FORMAT_yyyy_MM_dd);
              item.text = dateValue;
              item.updateValue(dateValue);
              update(item);
            }
          }, firstDate: item.firstDate, lastDate: item.lastDate);
        });
      }
      else if(item.maskedInputType == NUIFTextInputType.DOB){
        disableEdit = true;
        onWholeFieldTapAction = (){
          final selectedDate = NUIDTUtil.stringToDate(item.text ?? item.formValue?.value?.value);
          NUIDTUtil.dateSelect(context, selectedDate: selectedDate, selectListener: (date){
            if(date != null){
              final dateValue = dateToString(date, format: item.additionalFormat ?? NUIDTUtil.FORMAT_yyyy_MM_dd);
              item.text = dateValue;
              item.updateValue(dateValue);
              update(item);
            }
          }, firstDate: item.firstDate, lastDate: DateTime.now());
        };
        suffixIcon = IconButton(icon : item.suffixIcon ?? Icon(Icons.date_range, color: item.style.labelStyle.color), onPressed: (){
          final selectedDate = NUIDTUtil.stringToDate(item.text ?? item.formValue?.value?.value);
          NUIDTUtil.dateSelect(context, selectedDate: selectedDate, selectListener: (date){
            if(date != null){
              final dateValue = dateToString(date, format: item.additionalFormat ?? NUIDTUtil.FORMAT_yyyy_MM_dd);
              item.text = dateValue;
              item.updateValue(dateValue);
              update(item);
            }
          }, firstDate: item.firstDate, lastDate: DateTime.now());
        });
      }
      else if(item.maskedInputType == NUIFTextInputType.TIME){
        disableEdit = true;
        onWholeFieldTapAction = (){
          final selectedDate = NUIDTUtil.stringToDate(item.text ?? item.formValue?.value?.value);
          final selectedTime = selectedDate != null ? TimeOfDay.fromDateTime(selectedDate) : null;
          NUIDTUtil.timeSelect(context, initialTime: selectedTime, selectListener: (time){
            if(time != null){
              final timeFormat = item.additionalFormat;
              String timeValue;
              if(isNullOrEmpty(timeFormat) || match(timeFormat, "HH:mm")){
                timeValue = time.format(context);
              }
              else{
                timeValue = formatTime(context, time, use24HourFormat: false);
              }
              item.text = timeValue;
              item.updateValue(timeValue);
              update(item);
            }
          });
        };
        suffixIcon = IconButton(icon : item.suffixIcon ?? Icon(Icons.access_time_rounded, color: item.style.labelStyle.color), onPressed: (){
          final selectedDate = NUIDTUtil.stringToDate(item.text ?? item.formValue?.value?.value);
          final selectedTime = selectedDate != null ? TimeOfDay.fromDateTime(selectedDate) : null;
          NUIDTUtil.timeSelect(context, initialTime: selectedTime, selectListener: (time){
            if(time != null){
              final timeFormat = item.additionalFormat;
              String timeValue;
              if(isNullOrEmpty(timeFormat) || match(timeFormat, "HH:mm")){
                timeValue = time.format(context);
              }
              else{
                timeValue = formatTime(context, time, use24HourFormat: false);
              }
              item.text = timeValue;
              item.updateValue(timeValue);
              update(item);
            }
          });
        });
      }
      else if(item.maskedInputType == NUIFTextInputType.DATETIME){
        disableEdit = true;
        onWholeFieldTapAction = (){
          final selectedDate = NUIDTUtil.stringToDate(item.text ?? item.formValue?.value?.value);
          NUIDTUtil.dateSelect(context, selectedDate: selectedDate, selectListener: (date){
            if(date != null){
              NUIDTUtil.timeSelect(context, selectListener: (time){
                if(time != null) {
                  final dateTime = DateTime(date.year, date.month, date.day, time.hour, time.minute);
                  final dateValue = dateToString(dateTime, format: item.additionalFormat ?? NUIDTUtil.FORMAT_dd_MMM_yyyy_hh_mm_a);
                  item.text = dateValue;
                  item.updateValue(dateValue);
                  update(item);
                }
              });
            }
          }, firstDate: item.firstDate, lastDate: item.lastDate);
        };
        suffixIcon = IconButton(icon : item.suffixIcon ?? Icon(Icons.date_range, color: item.style.labelStyle.color), onPressed: (){
          final selectedDate = NUIDTUtil.stringToDate(item.text ?? item.formValue?.value?.value);
          NUIDTUtil.dateSelect(context, selectedDate: selectedDate, selectListener: (date){
            if(date != null){
              NUIDTUtil.timeSelect(context, selectListener: (time){
                if(time != null) {
                  final dateTime = DateTime(date.year, date.month, date.day, time.hour, time.minute);
                  final dateValue = dateToString(dateTime, format: item.additionalFormat ?? NUIDTUtil.FORMAT_dd_MMM_yyyy_hh_mm_a);
                  item.text = dateValue;
                  item.updateValue(dateValue);
                  update(item);
                }
              });
            }
          }, firstDate: item.firstDate, lastDate: item.lastDate);
        });
      }
      else{
        suffixIcon = item.suffixIcon;
        if(suffixIcon != null){
          //Add on tap listener to suffix icon
          suffixIcon = GestureDetector(
            child: suffixIcon,
            onTap: (){
              if(item.suffixOnTap != null) item.suffixOnTap();
            },
          );
        }
      }

      if(!NUIDataUtil.isNullOrEmpty(item.maskedInputType)){
        switch(item.maskedInputType){
          case NUIFTextInputType.SENTENCE : {
            item.inputType = TextInputType.text;
            item.textCapitalization = TextCapitalization.sentences;
            break;
          }
          case NUIFTextInputType.MULTILINE : {
            item.inputType = TextInputType.multiline;
            item.textCapitalization = TextCapitalization.sentences;
            item.textAlign = TextAlign.start;
            item.minLine = item.minLine ?? 10;
            break;
          }
          case NUIFTextInputType.EMAIL : {
            item.inputType = TextInputType.emailAddress;
            item.regExp = !isNullOrEmpty(item.regExp) ? item.regExp : StandardRegularExp.EMAILV2;
            item.maxLine = 1;
            item.textCapitalization = TextCapitalization.none;
            break;
          }
          case NUIFTextInputType.PASSWORD : {
            item.obscure = !_passwordToggle;
            item.inputType = TextInputType.visiblePassword;
            item.textCapitalization = TextCapitalization.none;
            item.maxLine = 1;
            break;
          }
          case NUIFTextInputType.PASSWORD_NEW : {
            item.obscure = !_passwordToggle;
            item.minLength = item.minLength ?? 8;
            item.inputType = TextInputType.visiblePassword;
            item.textCapitalization = TextCapitalization.none;
            item.maxLine = 1;
            item.contentCheck = (value){
              return TextAdditionalFormatting.passwordValidation(item, value);
            };
            break;
          }
          case NUIFTextInputType.NUMBER : {
            item.inputType = TextInputType.number;
            break;
          }
          case NUIFTextInputType.PHONE : {
            item.inputType = TextInputType.phone;
            item.maxLine = 1;
            break;
          }
          case NUIFTextInputType.DECIMAL : {
            item.inputType = TextInputType.number;
            break;
          }
          case NUIFTextInputType.NAME : {
            item.inputType = TextInputType.name;
            item.textCapitalization = TextCapitalization.words;
            break;
          }
          case NUIFTextInputType.MALAYSIAN_PHONE : {
            item.inputType = TextInputType.phone;
//            item.postTextFormatter = (value, inputType){
//
//            }
            break;
          }
          case NUIFTextInputType.MALAYSIAN_IC : {
            item.inputType = TextInputType.numberWithOptions(decimal: false);
            item.minLength = 14; //12 numbers with 2 '-'
            item.postTextFormatter = (value, inputType, handler){
              String text = value;
              if(text.length < _newText.length) { // handling backspace in keyboard
                _newText = text;
              }
              else if (!isNullOrEmpty(text) && !match(text, _newText)) { // handling typing new characters.
                String tempText = text.replaceAll("-", "");
                if(tempText.length == 6){
                  _newText = "$tempText-";
                }
                else if(tempText.length == 8){
                  _newText = "$text-";
                }
                else{
                  _newText = text;
                }
                if(!match(text, _newText)) {
                  _updateTextSelection(_newText, item.controller);
                }
              }
              _handler._newText = _newText;
              return _newText;
            };
            break;
          }
        }
      }

      final failedValidation = item.failed ?? false;

      decoration = InputDecoration(
          hintText: item.hint,
          hintStyle: item.style?.hintStyle,
          errorText: !failedValidation ? null : item.errorHelperText.isNullOrEmpty() ? null : item.errorHelperText,
          labelText: item.label,
          labelStyle: !failedValidation ? item.style?.labelStyle : item.style?.errorLabelStyle,
          icon: item.icon,
          prefix: item.prefix,
          prefixText: item.prefixText,
          prefixIcon: item.prefixIcon,
          suffix: item.suffix,
          suffixText: item.suffixText,
          alignLabelWithHint: true,
          suffixIcon: suffixIcon,
          errorMaxLines: 3,
          helperText: !failedValidation ? item.helperText : item.errorHelperText.isNullOrEmpty() ? null : item.errorHelperText,
          helperStyle: !failedValidation ? item.style?.helperStyle : item.style?.errorHelperStyle,
          errorStyle: item.style?.errorHelperStyle,
          isDense: item.style?.dense,
          contentPadding: item.style?.contentPadding,
          border: item.style?.design == NUIFTextDesign.BORDER ? item.getBorder() : null,
          enabledBorder: item.style?.design == NUIFTextDesign.BORDER ? item.getEnabledBorder() : null,
          disabledBorder: item.style?.design == NUIFTextDesign.BORDER ? item.getDisabledBorder() : null,
          focusedBorder: item.style?.design == NUIFTextDesign.BORDER ? item.getFocusedBorder() : null,
          errorBorder: item.style?.design == NUIFTextDesign.BORDER ? item.getErrorBorder() : null,
          focusedErrorBorder: item.style?.design == NUIFTextDesign.BORDER ? item.getErrorFocusedBorder() : null,
          fillColor: item.style?.fillColor
      );
    }
    else{
      decoration = InputDecoration.collapsed(
        hintText: item.hint,
        hintStyle: item.style?.hintStyle,
      );
    }

    item.controller?.text = (item.formValue?.value?.value == null ? item.text : item.formValue?.value?.value);
    if(item.controller != null) {
//      var cursorPos = item.controller.selection;
////      if (cursorPos.start > item.controller.text.length) {
//        cursorPos = new TextSelection.fromPosition(new TextPosition(offset: item.controller.text.length));
////      }
//      item.controller.selection = cursorPos;
      _updateTextSelection(item.controller?.text ?? "", item.controller);
    }

//    final minLines = item.maskedInputType == NUIFTextInputType.MULTILINE ? item.maxLine ?? 10 : 1;
    final inputFormatters = <TextInputFormatter>[];
    if(!isNullOrEmpty(item.blacklisting)){
      inputFormatters.add(FilteringTextInputFormatter.deny(item.blacklisting));
    }
    if(!isNullOrEmpty(item.whitelisting)){
      inputFormatters.add(FilteringTextInputFormatter.allow(item.whitelisting));
    }
    if(!isNullOrEmpty(item.inputFormatters)){
      inputFormatters.addAll(item.inputFormatters);
    }

    item.controller = item.controller ?? TextEditingController(text: item.formValue?.value?.value ?? item.text);
    final isMultiline = item.inputType == TextInputType.multiline;

    final textfield = TextField(
        controller: item.controller,
        onChanged: (value){
          if(item.onChanged != null) item.onChanged(value);
          if(item.validateOnChange){
            item.validate(false);
          }
        },
        onEditingComplete: (){
          if(item.onEditingComplete != null) item.onEditingComplete();
          if(item.validateOnEditComplete == true){
            item.validate(false);
          }
        },
        onSubmitted: (value){
          if(item.onSubmitted != null) item.onSubmitted(value);
          if(item.validateOnSubmit == true){
            item.validate(false);
          }
          if(item.inputAction == TextInputAction.done || item.inputAction == TextInputAction.next){
            focusNode.unfocus();
          }
        },
        autofocus: item.autoFocus,
        autocorrect: item.autoCorrect,
        focusNode: item.focusNode ?? focusNode,
        keyboardType: isMultiline == true ? TextInputType.multiline : item.inputType,
        textInputAction: isMultiline == true ? TextInputAction.newline : item.inputAction,
        textCapitalization: item.textCapitalization,
        enabled: item.enabled,
        readOnly: disableEdit,
        textAlign: item.textAlign,
        style: item.style?.textStyle,
        cursorColor: item.style?.cursor?.color ?? null,
        cursorRadius: item.style?.cursor?.cursorRadius ?? null,
        cursorWidth: item.style?.cursor?.cursorWidth ?? 2,
        maxLength: item.maxLength,
        maxLines: isMultiline == true ? 10 : item.maxLine ?? null,
        minLines: item.minLine ?? 1,
        obscureText: item.obscure,
        decoration: decoration,
        inputFormatters: inputFormatters,
    );

    if(onWholeFieldTapAction != null){
      return GestureDetector(
        onTap: (){
          logNUI("NUIFText", "Calling on tap on whole text field");
          onWholeFieldTapAction();
        },
        child: Container(
          color: Colors.transparent,
          child: IgnorePointer(
            child: textfield
          ),
        ),
      );
    }

    return textfield;
  }

  @override
  void dispose() {
    // item.controller?.dispose();
    _handler._updateState = null;
    _handler._updateText = null;
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    return item.wrapWidget(_bindWithTextItem(item));
  }
}
